# Copyright 2018 The GNOME Music developers
#
# GNOME Music is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from gi.repository import Gtk

from gnomemusic.scrobbler import GoaLastFM


@Gtk.Template(resource_path="/org/gnome/Music/ui/AppMenu.ui")
class AppMenu(Gtk.Popover):
    """AppMenu shown from the HeaderBar within the main view"""

    __gtype_name__ = "AppMenu"

    _lastfm_box = Gtk.Template.Child()
    _lastfm_switch = Gtk.Template.Child()

    def __init__(self, application):
        """Initialize the application menu

        :param Application application: Application object
        """
        super().__init__()

        self._lastfm_switcher_id = None

        self._lastfm_configure_action = application.lookup_action(
            "lastfm-configure")

        self._lastfm_scrobbler = application.props.lastfm_scrobbler
        self._lastfm_scrobbler.connect(
            "notify::can-scrobble", self._on_scrobbler_state_changed)
        self._on_scrobbler_state_changed(None, None)

    def _on_scrobbler_state_changed(self, klass, args):
        state = self._lastfm_scrobbler.props.account_state

        if state == GoaLastFM.State.NOT_AVAILABLE:
            self._lastfm_configure_action.props.enabled = False
            return

        self._lastfm_configure_action.props.enabled = True

        if state == GoaLastFM.State.NOT_CONFIGURED:
            self._lastfm_box.props.visible = False
            if self._lastfm_switcher_id is not None:
                self._lastfm_switch.disconnect(self._lastfm_switcher_id)
                self._lastfm_switcher_id = None
            return

        self._lastfm_box.props.visible = True
        if self._lastfm_switcher_id is None:
            self._lastfm_switcher_id = self._lastfm_switch.connect(
                "state-set", self._on_lastfm_switch_active)

        with self._lastfm_switch.handler_block(self._lastfm_switcher_id):
            can_scrobble = self._lastfm_scrobbler.props.can_scrobble
            self._lastfm_switch.props.state = can_scrobble

    def _on_lastfm_switch_active(self, klass, state):
        self._lastfm_scrobbler.props.can_scrobble = state



# Copyright 2022 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Optional

from gi.repository import Gdk, Gio, GLib, GObject

from gnomemusic.musiclogger import MusicLogger


class MediaArtLoader(GObject.GObject):
    """Handles retrieval of MediaArt cache art

    Signals when the media is loaded and passes a texture or None.
    """

    __gtype_name__ = "MediaArtLoader"

    __gsignals__ = {
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (object, ))
    }

    _chunksize = 32768
    _log = MusicLogger()

    def __init__(self) -> None:
        """Intialize MediaArtLoader
        """
        super().__init__()

        self._bytearray = bytearray()
        self._texture: Optional[Gdk.Texture] = None

    def start(self, uri: str) -> None:
        """Start the cache query

        :param str uri: The MediaArt uri
        """
        thumb_file = Gio.File.new_for_uri(uri)
        if thumb_file:
            thumb_file.read_async(
                GLib.PRIORITY_DEFAULT_IDLE, None, self._open_stream)
        else:
            self.emit("finished", self._texture)

    def _open_stream(
            self, thumb_file: Gio.File, result: Gio.AsyncResult) -> None:
        try:
            stream = thumb_file.read_finish(result)
        except GLib.Error as error:
            self._log.warning(
                "Error: {}, {}".format(error.domain, error.message))
            self.emit("finished", self._texture)
        else:
            stream.read_bytes_async(
                self._chunksize, GLib.PRIORITY_DEFAULT_IDLE, None,
                self._read_bytes_async_cb)

    def _read_bytes_async_cb(
            self, stream: Gio.FileInputStream,
            result: Gio.AsyncResult) -> None:
        try:
            gbytes = stream.read_bytes_finish(result)
        except GLib.Error as error:
            self._log.warning(
                "Error: {}, {}".format(error.domain, error.message))
            stream.close_async(
                GLib.PRIORITY_DEFAULT_IDLE, None, self._close_stream)
            return

        gbytes_size = gbytes.get_size()
        if gbytes_size > 0:
            self._bytearray += gbytes.unref_to_data()

            stream.read_bytes_async(
                self._chunksize, GLib.PRIORITY_DEFAULT_IDLE, None,
                self._read_bytes_async_cb)
        else:
            # FIXME: Use GTask to load textures async.
            # See pygobject#114 for bytes conversion.
            self._texture = Gdk.Texture.new_from_bytes(
                GLib.Bytes(bytes(self._bytearray)))

            self._bytearray = bytearray()

            stream.close_async(
                GLib.PRIORITY_DEFAULT_IDLE, None, self._close_stream)

    def _close_stream(
            self, stream: Gio.InputStream, result: Gio.AsyncResult) -> None:
        try:
            stream.close_finish(result)
        except GLib.Error as error:
            self._log.warning(
                "Error: {}, {}".format(error.domain, error.message))

        self.emit("finished", self._texture)



# Copyright 2021 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from gi.repository import GLib, GObject, Gio


class FileExistsAsync(GObject.GObject):
    """Gio.File.file_exists async variant
    """

    __gtype_name__ = "FileExistsAsync"

    __gsignals__ = {
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (bool, ))
    }

    def __init__(self) -> None:
        """Initialize FileExistsAsync
        """
        super().__init__()

    def start(self, thumb_file: Gio.File) -> None:
        """Start async file_exists lookup

        :param Gio.File thumb_file: File to check
        """
        thumb_file.query_info_async(
            Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE,
            GLib.PRIORITY_DEFAULT_IDLE, None, self._on_query_info_finished)

    def _on_query_info_finished(
            self, thumb_file: Gio.File, res: Gio.AsyncResult) -> None:
        exists = True
        try:
            thumb_file.query_info_finish(res)
        except GLib.Error:
            # This indicates that the file has not been created, so
            # there is no art in the MediaArt cache.
            exists = False

        self.emit("finished", exists)



# Copyright 2020 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

import gi
gi.require_version("MediaArt", "2.0")
from gi.repository import GObject, MediaArt

from gnomemusic.asyncqueue import AsyncQueue
from gnomemusic.embeddedart import EmbeddedArt
from gnomemusic.fileexistsasync import FileExistsAsync
from gnomemusic.griloartqueue import GriloArtQueue
from gnomemusic.utils import CoreObjectType


class AlbumArt(GObject.GObject):
    """AlbumArt retrieval object
    """

    _async_queue = AsyncQueue("AlbumArt")

    def __init__(self, application, corealbum):
        """Initialize AlbumArt

        :param Application application: The application object
        :param CoreAlbum corealbum: The corealbum to use
        """
        super().__init__()

        self._corealbum = corealbum
        self._coregrilo = application.props.coregrilo
        self._album = self._corealbum.props.title
        self._artist = self._corealbum.props.artist

        self._grilo_art_queue = GriloArtQueue(application)

        self._in_cache()

    def _on_embedded_art_found(self, embeddedart, found):
        if found:
            self._in_cache()
        else:
            self._grilo_art_queue.queue(self._corealbum, CoreObjectType.ALBUM)

    def _in_cache(self):
        success, thumb_file = MediaArt.get_file(
            self._artist, self._album, "album")
        if not success:
            return

        def on_file_exists_async_finished(obj, result):
            if result:
                self._corealbum.props.thumbnail = thumb_file.get_uri()
            else:
                embedded = EmbeddedArt()
                embedded.connect("art-found", self._on_embedded_art_found)
                embedded.query(self._corealbum, self._album)

        file_exists_async = FileExistsAsync()
        file_exists_async.connect(
            "finished", on_file_exists_async_finished)
        self._async_queue.queue(file_exists_async, thumb_file)



# Copyright 2021 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Any, Optional, Union, cast
import typing

from gi.repository import Gio, GObject, Gtk

from gnomemusic.grilowrappers.grltrackerplaylists import Playlist
from gnomemusic.songtoast import SongToast
from gnomemusic.widgets.playlistdialog import PlaylistDialog
from gnomemusic.widgets.songwidget import SongWidget
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coresong import CoreSong
    CoreObject = Union[CoreAlbum, CoreSong, Playlist]


@Gtk.Template(resource_path="/org/gnome/Music/ui/SongWidgetMenu.ui")
class SongWidgetMenu(Gtk.PopoverMenu):

    __gtype_name__ = "SongWidgetMenu"

    def __init__(
            self, application: Application,
            song_widget: Union[SongWidget, Gtk.ListItem],
            coreobject: Optional[CoreObject]) -> None:
        """Menu to interact with the song of a SongWidget.

        :param Application application: The application object
        :param SongWidget song_widget: The songwidget associated with the menu
        :param Union[CoreAlbum, CoreSong, Playlist] coreobject: The
            coreobject associated with the menu
        """
        super().__init__()

        self._application = application
        self._coremodel = application.props.coremodel
        self._player = application.props.player
        self._window = application.props.window

        self._coreobject = coreobject
        self._coresong: CoreSong
        self._song_widget = song_widget
        self.props.coreobject = coreobject

        self._playlist_dialog: Optional[PlaylistDialog] = None

        action_group = Gio.SimpleActionGroup()
        action_entries = [
            ("play", self._play_song),
            ("add_playlist", self._add_to_playlist)
        ]
        if (isinstance(self._coreobject, Playlist)
                and not self._coreobject.props.is_smart):
            action_entries.append(
                ("remove_playlist", self._remove_from_playlist))
        elif not isinstance(self._coreobject, Playlist):
            self.props.menu_model.remove(2)

        for name, callback in action_entries:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", callback)
            action_group.add_action(action)

        self.insert_action_group("songwidget", action_group)

    def _play_song(self, action: Gio.Simple, param: Any) -> None:
        self.popdown()
        signal_id = 0

        def _on_playlist_loaded(klass, playlist_type):
            self._player.play(self._coresong)
            self._coremodel.disconnect(signal_id)

        signal_id = self._coremodel.connect(
            "playlist-loaded", _on_playlist_loaded)
        self._coremodel.props.active_core_object = self._coreobject

    def _add_to_playlist(self, action: Gio.Simple, param: Any) -> None:

        def on_response(dialog: PlaylistDialog, response_id: int) -> None:
            if not self._playlist_dialog:
                return

            if response_id == Gtk.ResponseType.ACCEPT:
                playlist = self._playlist_dialog.props.selected_playlist
                playlist.add_songs([self._coresong])

            self._playlist_dialog.destroy()
            self._playlist_dialog = None

        self.popdown()
        self._playlist_dialog = PlaylistDialog(self._application)
        self._playlist_dialog.props.transient_for = self._window
        self._playlist_dialog.connect("response", on_response)
        self._playlist_dialog.present()

    def _remove_from_playlist(self, action: Gio.Simple, param: Any) -> None:
        self.popdown()
        position = self._song_widget.get_index()
        SongToast(
            self._application, cast(Playlist, self._coreobject), position,
            self._coresong)

    @GObject.Property(type=GObject.GObject)
    def coreobject(self) -> CoreObject:
        return self._coreobject

    @coreobject.setter  # type: ignore
    def coreobject(self, coreobject: CoreObject) -> None:
        self._coreobject = coreobject

        if isinstance(self._song_widget, SongWidget):
            self._coresong = self._song_widget.props.coresong
        else:
            self._coresong = coreobject



# Copyright 2022 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from enum import IntEnum
from typing import Dict, Optional, Tuple, Union
import time
import typing

from gi.repository import GLib, GObject, Gdk, Gio

from gnomemusic.asyncqueue import AsyncQueue
from gnomemusic.musiclogger import MusicLogger
from gnomemusic.mediaartloader import MediaArtLoader
if typing.TYPE_CHECKING:
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coreartist import CoreArtist
    from gnomemusic.coresong import CoreSong

if typing.TYPE_CHECKING:
    CoreObject = Union[CoreAlbum, CoreArtist, CoreSong]


class TextureCache(GObject.GObject):
    """Retrieval and cache for artwork textures
    """

    class LoadingState(IntEnum):
        """The loading status of the URI

        AVAILABLE: The texture is currently cached
        UNAVAILABLE: No texture is available for the URI
        CLEARED: The texture was available, has been cleared and
            should be available on lookup
        """
        AVAILABLE = 0
        UNAVAILABLE = 1
        CLEARED = 2

    __gtype_name__ = "TextureCache"

    __gsignals__ = {
        "texture": (GObject.SignalFlags.RUN_FIRST, None, (object, ))
    }

    # Music has two main cycling views (AlbumsView and ArtistsView),
    # both have around 200 cycling items each when fully used. For
    # the cache to be useful it needs to be larger than the given
    # numbers combined.
    _MAX_CACHE_SIZE = 800

    _async_queue = AsyncQueue("TextureCache")
    _cleanup_id = 0
    _log = MusicLogger()
    _memory_monitor = Gio.MemoryMonitor.dup_default()
    _size = _MAX_CACHE_SIZE
    _textures: Dict[str, Tuple[
        TextureCache.LoadingState, float, Optional[Gdk.Texture]]] = {}

    def __init__(self) -> None:
        """Initialize Texturecache
        """
        super().__init__()

        self._art_loader: MediaArtLoader
        self._art_loading_id = 0

        if TextureCache._cleanup_id == 0:
            TextureCache._cleanup_id = GLib.timeout_add_seconds(
                10, TextureCache._cache_cleanup)
            TextureCache._memory_monitor.connect(
                "low-memory-warning", TextureCache._low_memory_warning)

    def clear_pending_lookup_callback(self) -> None:
        """Disconnect ongoing lookup callback
        """
        if self._art_loading_id != 0:
            self._art_loader.disconnect(self._art_loading_id)
            self._art_loading_id = 0

    def lookup(self, uri: str) -> None:
        """Look up a texture for the given MediaArt uri

        :param str uri: The MediaArt uri
        """
        self.clear_pending_lookup_callback()

        if uri in TextureCache._textures.keys():
            state, _, texture = TextureCache._textures[uri]
            if state in [
                    TextureCache.LoadingState.AVAILABLE,
                    TextureCache.LoadingState.UNAVAILABLE]:
                self.emit("texture", texture)
                TextureCache._textures[uri] = (state, time.time(), texture)
                return

        self._art_loader = MediaArtLoader()
        self._art_loading_id = self._art_loader.connect(
            "finished", self._on_art_loading_finished, uri)
        self._async_queue.queue(self._art_loader, uri)

    @classmethod
    def _low_memory_warning(
            cls, mm: Gio.MemoryMonitor,
            level: Gio.MemoryMonitorWarningLevel) -> None:
        if level < Gio.MemoryMonitorWarningLevel.LOW:
            TextureCache._size = TextureCache._MAX_CACHE_SIZE
        else:
            # List slicing with 0 gives an empty list in
            # _cache_cleanup.
            TextureCache._size = 1

    @classmethod
    def _cache_cleanup(cls) -> None:
        """Sorts the available cache entries by recency and evicts
        the oldest items to match the maximum cache size.
        """
        sorted_available = {
            k: (state, t, texture)
            for k, (state, t, texture) in sorted(
                TextureCache._textures.items(), key=lambda item: item[1][1])
            if state in [TextureCache.LoadingState.AVAILABLE]}

        sorted_available_l = len(sorted_available)
        if sorted_available_l < TextureCache._size:
            return GLib.SOURCE_CONTINUE

        keys_to_clear = list(sorted_available.keys())[:-TextureCache._size]
        for key in keys_to_clear:
            state, t, texture = TextureCache._textures[key]
            TextureCache._textures[key] = (
                TextureCache.LoadingState.CLEARED, t, None)

        keys_l = len(keys_to_clear)
        TextureCache._log.info(
            f"Cleared {keys_l} items, texture cache contains"
            f" {sorted_available_l-keys_l} available items.")

        return GLib.SOURCE_CONTINUE

    def _on_art_loading_finished(
            self, art_loader: MediaArtLoader, texture: Gdk.Texture,
            uri: str) -> None:
        if texture:
            state = TextureCache.LoadingState.AVAILABLE
        else:
            state = TextureCache.LoadingState.UNAVAILABLE

        TextureCache._textures[uri] = (state, time.time(), texture)

        self.emit("texture", texture)



# Copyright 2019 The GNOME Music Developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
import typing

from gettext import gettext as _
from gi.repository import GObject, Gtk
from typing import Dict, List

from gnomemusic.coverpaintable import CoverPaintable
from gnomemusic.utils import ArtSize, DefaultIconType
from gnomemusic.widgets.headerbar import HeaderBar
from gnomemusic.widgets.albumwidget import AlbumWidget
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.corealbum import CoreAlbum


@Gtk.Template(resource_path="/org/gnome/Music/ui/AlbumsView.ui")
class AlbumsView(Gtk.Stack):
    """Gridlike view of all albums

    Album activation switches to AlbumWidget.
    """

    __gtype_name__ = "AlbumsView"

    icon_name = GObject.Property(
        type=str, default="media-optical-cd-audio-symbolic",
        flags=GObject.ParamFlags.READABLE)
    search_mode_active = GObject.Property(type=bool, default=False)
    selection_mode = GObject.Property(type=bool, default=False)
    title = GObject.Property(
        type=str, default=_("Albums"), flags=GObject.ParamFlags.READABLE)

    _album_scrolled_window = Gtk.Template.Child()
    _scrolled_window = Gtk.Template.Child()
    _gridview = Gtk.Template.Child()

    def __init__(self, application: Application) -> None:
        """Initialize AlbumsView

        :param application: The Application object
        """
        super().__init__(transition_type=Gtk.StackTransitionType.CROSSFADE)

        self.props.name = "albums"

        self._application = application
        self._window = application.props.window
        self._headerbar = self._window.props.headerbar

        self._list_item_bindings: Dict[
            Gtk.ListItem, List[GObject.Binding]] = {}
        self._list_item_star_controllers: Dict[
            Gtk.ListItem, List[GObject.Binding]] = {}

        list_item_factory = Gtk.SignalListItemFactory()
        list_item_factory.connect("setup", self._setup_list_item)
        list_item_factory.connect("bind", self._bind_list_item)
        list_item_factory.connect("unbind", self._unbind_list_item)

        self._gridview.props.factory = list_item_factory

        self._selection_model = Gtk.MultiSelection.new(
            self._application.props.coremodel.props.albums_sort)
        self._gridview.props.model = self._selection_model

        self._gridview.connect("activate", self._on_album_activated)

        self.bind_property(
            "selection-mode", self._gridview, "single-click-activate",
            GObject.BindingFlags.SYNC_CREATE
            | GObject.BindingFlags.INVERT_BOOLEAN)
        self.bind_property(
            "selection-mode", self._gridview, "enable-rubberband",
            GObject.BindingFlags.SYNC_CREATE)
        self.bind_property(
            "selection-mode", self._window, "selection-mode",
            GObject.BindingFlags.DEFAULT)

        self._window.connect(
            "notify::selection-mode", self._on_selection_mode_changed)

        self._album_widget = AlbumWidget(self._application)
        self._album_widget.bind_property(
            "selection-mode", self, "selection-mode",
            GObject.BindingFlags.BIDIRECTIONAL)

        viewport = self._album_scrolled_window.get_first_child()
        viewport.set_child(self._album_widget)

        self.connect(
            "notify::search-mode-active", self._on_search_mode_changed)

    def _on_selection_mode_changed(self, widget, data=None):
        selection_mode = self._window.props.selection_mode
        if (selection_mode == self.props.selection_mode
                or self.get_parent().get_visible_child() != self):
            return

        self.props.selection_mode = selection_mode
        if not self.props.selection_mode:
            self.deselect_all()

    def _on_search_mode_changed(self, klass, param):
        if (not self.props.search_mode_active
                and self._headerbar.props.stack.props.visible_child == self
                and self.get_visible_child_name() == "widget"):
            self._set_album_headerbar(self._album_widget.props.corealbum)

    def _back_button_clicked(self, widget, data=None):
        self._headerbar.state = HeaderBar.State.MAIN
        self.props.visible_child = self._scrolled_window

    def _on_album_activated(self, widget, position):
        corealbum = widget.props.model[position]

        self._album_widget.props.corealbum = corealbum
        self._set_album_headerbar(corealbum)
        self.props.visible_child = self._album_scrolled_window

    def _on_child_activated(self, widget, child, user_data=None):
        corealbum = child.props.corealbum
        if self.props.selection_mode:
            return

        # Update and display the album widget if not in selection mode
        self._album_widget.props.corealbum = corealbum

        self._set_album_headerbar(corealbum)
        self.set_visible_child_name("widget")

    def _set_album_headerbar(self, corealbum: CoreAlbum) -> None:
        self._headerbar.props.state = HeaderBar.State.CHILD
        self._headerbar.set_label_title(
            corealbum.props.title, corealbum.props.artist)

    def _toggle_all_selection(self, selected):
        """Selects or deselects all items.
        """
        with self._application.props.coreselection.freeze_notify():
            if self.get_visible_child_name() == "widget":
                if selected is True:
                    self._album_widget.select_all()
                else:
                    self._album_widget.deselect_all()
            else:
                if selected:
                    self._selection_model.select_all()
                else:
                    self._selection_model.unselect_all()

    def select_all(self):
        self._toggle_all_selection(True)

    def deselect_all(self):
        self._toggle_all_selection(False)

    def _setup_list_item(
            self, factory: Gtk.SignalListItemFactory,
            list_item: Gtk.ListItem) -> None:
        builder = Gtk.Builder.new_from_resource(
            "/org/gnome/Music/ui/AlbumCoverListItem.ui")
        list_item.props.child = builder.get_object("_album_cover")
        cover_image = list_item.props.child.get_first_child().get_first_child()
        cover_image.set_size_request(
            ArtSize.MEDIUM.width, ArtSize.MEDIUM.height)
        cover_image.props.paintable = CoverPaintable(
            self, ArtSize.MEDIUM, icon_type=DefaultIconType.ALBUM)

        self.bind_property(
            "selection-mode", list_item, "selectable",
            GObject.BindingFlags.SYNC_CREATE)
        self.bind_property(
            "selection-mode", list_item, "activatable",
            GObject.BindingFlags.SYNC_CREATE
            | GObject.BindingFlags.INVERT_BOOLEAN)

    def _bind_list_item(
            self, factory: Gtk.SignalListItemFactory,
            list_item: Gtk.ListItem) -> None:
        album_cover = list_item.props.child
        corealbum = list_item.props.item

        cover_image = album_cover.get_first_child().get_first_child()
        check = cover_image.get_next_sibling()
        album_label = album_cover.get_first_child().get_next_sibling()
        artist_label = album_label.get_next_sibling()

        b1 = corealbum.bind_property(
            "corealbum", cover_image.props.paintable, "coreobject",
            GObject.BindingFlags.SYNC_CREATE)
        b2 = corealbum.bind_property(
            "title", album_label, "label", GObject.BindingFlags.SYNC_CREATE)
        b3 = corealbum.bind_property(
            "artist", artist_label, "label", GObject.BindingFlags.SYNC_CREATE)

        b4 = list_item.bind_property(
            "selected", corealbum, "selected",
            GObject.BindingFlags.SYNC_CREATE)
        b5 = list_item.bind_property(
            "selected", check, "active", GObject.BindingFlags.SYNC_CREATE)
        b6 = self.bind_property(
            "selection-mode", check, "visible",
            GObject.BindingFlags.SYNC_CREATE)

        def on_activated(widget, value):
            if check.props.active:
                self._selection_model.select_item(
                    list_item.get_position(), False)
            else:
                self._selection_model.unselect_item(
                    list_item.get_position())

        # the listitem selected property is read-only.
        # It cannot be bound from the check active property.
        # It is necessary to update the selection model in order
        # to update it.
        check.connect("notify::active", on_activated)

        self._list_item_bindings[list_item] = [b1, b2, b3, b4, b5, b6]

    def _unbind_list_item(
            self, factory: Gtk.SignalListItemFactory,
            list_item: Gtk.ListItem) -> None:
        bindings = self._list_item_bindings.pop(list_item)
        [binding.unbind() for binding in bindings]

        album_cover = list_item.props.child

        cover_image = album_cover.get_first_child().get_first_child()
        check = cover_image.get_next_sibling()

        signal_id, detail_id = GObject.signal_parse_name(
            "notify::active", check, True)
        handler_id = GObject.signal_handler_find(
            check, GObject.SignalMatchType.ID, signal_id, detail_id, None, 0,
            0)
        check.disconnect(handler_id)



# Copyright 2020 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

import gi
gi.require_versions({"GstPbutils": "1.0", "GstTag": "1.0", "MediaArt": "2.0"})
from gi.repository import (
    GLib, GObject, MediaArt, GdkPixbuf, Gio, Gst, GstTag, GstPbutils)

from gnomemusic.musiclogger import MusicLogger


class EmbeddedArt(GObject.GObject):
    """Lookup local art

    1. Embedded art using GStreamer
    2. Available in the directory using MediaArt
    """

    _log = MusicLogger()

    __gsignals__ = {
        "art-found": (GObject.SignalFlags.RUN_FIRST, None, (bool, ))
    }

    def __init__(self):
        """Initialize EmbeddedArt
        """
        super().__init__()

        try:
            Gst.init(None)
            GstPbutils.pb_utils_init()
        except GLib.Error as error:
            self._log.warning(
                "Error: {}, {}".format(error.domain, error.message))
            return

        self._media_art = MediaArt.Process.new()

        self._album = None
        self._artist = None
        self._coreobject = None
        self._file = None

    def query(self, coreobject, title):
        """Start the local query

        :param coreobject: The CoreAlbum or CoreSong to search art for
        :param str title: The album title for the CoreAlbum or CoreSong
        """
        self._album = title
        self._artist = coreobject.props.artist
        self._coreobject = coreobject

        try:
            if coreobject.props.url is None:
                self.emit("art-found", False)
                return
        except AttributeError:
            self.emit("art-found", False)
            return

        try:
            discoverer = GstPbutils.Discoverer.new(Gst.SECOND)
        except GLib.Error as error:
            self._log.warning(
                "Error: {}, {}".format(error.domain, error.message))
            self._lookup_cover_in_directory()
            return

        discoverer.connect("discovered", self._discovered)
        discoverer.start()

        success, file = MediaArt.get_file(self._artist, self._album, "album")

        if not success:
            self.emit("art-found", False)
            discoverer.stop()
            return

        self._file = file

        success = discoverer.discover_uri_async(self._coreobject.props.url)

        if not success:
            self._log.warning("Could not add url to discoverer.")
            self.emit("art-found", False)
            discoverer.stop()
            return

    def _discovered(self, discoverer, info, error):
        tags = info.get_tags()
        index = 0

        if (error is not None
                or tags is None):
            if error:
                self._log.warning("Discoverer error: {}, {}".format(
                    Gst.CoreError(error.code), error.message))
            discoverer.stop()
            self.emit("art-found", False)
            return

        while True:
            success, sample = tags.get_sample_index(Gst.TAG_IMAGE, index)
            if not success:
                break
            index += 1
            struct = sample.get_info()
            if struct is None:
                break
            success, image_type = struct.get_enum(
                "image-type", GstTag.TagImageType)
            if not success:
                continue
            if image_type not in [
                    GstTag.TagImageType.UNDEFINED,
                    GstTag.TagImageType.FRONT_COVER]:
                continue

            buf = sample.get_buffer()
            success, map_info = buf.map(Gst.MapFlags.READ)
            if not success:
                continue

            istream = Gio.MemoryInputStream.new_from_data(map_info.data)
            GdkPixbuf.Pixbuf.new_from_stream_async(
                istream, None, self._pixbuf_from_stream_finished)
            discoverer.stop()
            return

        discoverer.stop()
        self._lookup_cover_in_directory()

    def _pixbuf_from_stream_finished(
            self, stream: Gio.MemoryInputStream,
            result: Gio.AsyncResult) -> None:
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_stream_finish(result)
        except GLib.Error as error:
            self._log.warning(f"Error: {error.domain}, {error.message}")
            self._lookup_cover_in_directory()
        else:
            self._file.create_async(
                Gio.FileCreateFlags.NONE, GLib.PRIORITY_DEFAULT_IDLE, None,
                self._output_stream_created, pixbuf)
        finally:
            stream.close_async(
                GLib.PRIORITY_DEFAULT_IDLE, None, self._stream_closed)

    def _output_stream_created(
            self, stream: Gio.FileOutputStream, result: Gio.AsyncResult,
            pixbuf: GdkPixbuf.Pixbuf) -> None:
        try:
            output_stream = stream.create_finish(result)
        except GLib.Error as error:
            # File already exists.
            self._log.info(f"Error: {error.domain}, {error.message}")
            self.emit("art-found", True)
        else:
            pixbuf.save_to_streamv_async(
                output_stream, "jpeg", None, None, None,
                self._output_stream_saved, output_stream)

    def _output_stream_saved(
            self, pixbuf: GdkPixbuf.Pixbuf, result: Gio.AsyncResult,
            output_stream: Gio.FileOutputStream) -> None:
        try:
            pixbuf.save_to_stream_finish(result)
        except GLib.Error as error:
            self._log.warning(f"Error: {error.domain}, {error.message}")
            self._lookup_cover_in_directory()
        else:
            self.emit("art-found", True)
        finally:
            output_stream.close_async(
                GLib.PRIORITY_DEFAULT_IDLE, None, self._stream_closed)

    def _stream_closed(
            self, stream: Gio.OutputStream, result: Gio.AsyncResult) -> None:
        try:
            stream.close_finish(result)
        except GLib.Error as error:
            self._log.warning(f"Error: {error.domain}, {error.message}")

    def _lookup_cover_in_directory(self):
        # Find local art in cover.jpeg files.
        self._media_art.uri_async(
            MediaArt.Type.ALBUM, MediaArt.ProcessFlags.NONE,
            self._coreobject.props.url, self._artist, self._album,
            GLib.PRIORITY_DEFAULT_IDLE, None, self._uri_async_cb, None)

    def _uri_async_cb(self, src, result, data):
        try:
            success = self._media_art.uri_finish(result)
            if success:
                self.emit("art-found", True)
                return
        except GLib.Error as error:
            if MediaArt.Error(error.code) == MediaArt.Error.SYMLINK_FAILED:
                # This error indicates that the coverart has already
                # been linked by another concurrent lookup.
                self.emit("art-found", True)
                return
            else:
                self._log.warning("Error: {}, {}".format(
                    MediaArt.Error(error.code), error.message))

        self.emit("art-found", False)
